import java.applet.Applet;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.image.IndexColorModel;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.util.Arrays;
import java.util.Random;

public class Fire3 extends Applet implements Runnable {
    final Random rand = new Random();
    final Object paintLock = new Object();
    BufferedImage flame;
    int start;
    int[] count;
    int[] pixels;
    int[] weight;
    public void filter(Raster src, WritableRaster dest) {
        final int width = src.getWidth();
        final int height = src.getHeight();
        if (start < width+1) {
            start = width+1;
        }
        if (pixels == null || pixels.length < width*height) {
            pixels = new int[height * width];
            count = new int[pixels.length];
            weight = new int[pixels.length];
        } else {
            Arrays.fill(count, 0);
        }
        Arrays.fill(weight, 10);
        src.getPixels(0, 0, width, height, pixels);
        for (int i = start; i < pixels.length-1; ++i) {
            if (pixels[i] > 0) {
                final int left = i - 1;
                final int up = i - width;
                final int upleft = left - width;
                //final int right = i + 1;
                final int weightedPixel1 = pixels[i];
                final int weightedPixel4 = pixels[i] * 4;

                count[i] += weightedPixel1;
                count[left] += weightedPixel1;
                if (weightedPixel1 < 32)
                {
                    count[up] += weightedPixel4;
                    count[upleft] += weightedPixel4;
                } else {
                    final int up2 = up - width;
                    final int up2left = up2 - 1;
                    if (weightedPixel1 < 64) {
                        final int weightedPixel3 = pixels[i] * 3;
                        count[up] += weightedPixel3;
                        count[upleft] += weightedPixel3;
                        count[up2] += weightedPixel1;
                        count[up2left] += weightedPixel1;
                        --weight[up];
                        --weight[upleft];
                        ++weight[up2];
                        ++weight[up2left];
                    } else {
                        final int weightedPixel2 = pixels[i] * 2;
                        final int up3 = up2 - width;
                        final int up3left = up3 - 1;
                        count[up] += weightedPixel2;
                        count[upleft] += weightedPixel2;
                        count[up2] += weightedPixel1;
                        count[up2left] += weightedPixel1;
                        count[up3] += weightedPixel1;
                        count[up3left] += weightedPixel1;
                        weight[up] -= 2;
                        weight[upleft] -= 2;
                        ++weight[up2];
                        ++weight[up2left];
                        ++weight[up3];
                        ++weight[up3left];
                        if (weightedPixel1 >=  96) {
                            final int up4 = up3 - width;
                            final int up4left = up3 - 1;
                            final int up4left2 = up4 + 1;
                            count[up4] += weightedPixel1;
                            count[up4left] += weightedPixel1;
                            count[up4left2] += weightedPixel1;
                            ++weight[up4];
                            ++weight[up4left];
                            ++weight[up4left2];
                        }
                    }
                }
            }
        }
        int newStart = pixels.length;
        for (int i = pixels.length - (width+1) - 1; i >= start; --i) {
            if (count[i] > 0) {
                count[i] /= weight[i];
                if (count[i] > 0) {
                    count[i] -= 1;
                    newStart = i;
                }
            }
        }
        start = newStart - width*4;
        dest.setPixels(0, 0, width, height-1, count);
    }

    final IndexColorModel indexColorModel;

    public Fire3() {
        byte[] r = new byte[256], g = new byte[256], b = new byte[256];
        for (int i = 0; i < 64; ++i) {
            b[i + 128] = g[i + 64] = r[i] = (byte)(i * 4);
            b[i + 192] = g[i + 128] = r[i+128] = r[i+64] = (byte)255;
            b[i + 64] = b[i] = g[i] = 0;
            r[i + 192] = g[i + 192] = (byte) ((byte) 255 - (r[i + 192] = (byte) (i*4)));
         }
        indexColorModel = new IndexColorModel(8, r.length, r, g, b);
    }

    volatile boolean notDone = true;

    public void resize() {
        synchronized(paintLock) {
            flame = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_BYTE_INDEXED, indexColorModel);
        }
    }

    public void init() {
        resize();
    }

    public void start() {
        synchronized(paintLock) {
            new Thread(this).start();
            notDone = true;
        }
    }

    public void run() {
        synchronized(paintLock) {
            while (notDone) {
                step();
                repaint();
                try {
                    paintLock.wait();
                } catch (InterruptedException e) {
                }
            }
        }
    }

    public void stop() {
        synchronized(paintLock) {
            notDone = false;
            paintLock.notifyAll();
        }
    }


    public void step() {
        WritableRaster flameRaster = flame.getRaster();
        final int[] hot = {0xFF};
        final int[] cold = {0x00};
        for (int i = 0; i < flame.getWidth()/16; ++i) {
            //flameRaster.setPixel(rand.nextInt(flame.getWidth()), flame.getHeight() -1, hot);
            flameRaster.setPixel(rand.nextInt(flame.getWidth()), flame.getHeight() -1, hot);
            flameRaster.setPixel(rand.nextInt(flame.getWidth()), flame.getHeight() -1, cold);
        }
        filter(flame.getData(), flame.getRaster());
    }

    public void update(Graphics g) {
        paint(g);
    }

    public void paint(Graphics g) {
        synchronized(paintLock) {
            g.drawImage(flame, 0, 0, flame.getWidth(), flame.getHeight()-2,0, 0, flame.getWidth(), flame.getHeight()-2, null);
            paintLock.notifyAll();
        }
    }
}
